#include <stdio.h>
#include <malloc.h>
#include "structs.h"
#include "protos.h"

struct s_diagram *diagram;
struct s_object **objarray;
int nobjects;
int curry;

int barwidth = 16;
int eventheight = 20;
int textspacing = 3;	/* amount of space under arrow descriptions */
int defobjwidth = 80;
int objheight = 24;

void drawarrow(struct s_object *src, struct s_object *dest,
		int dotflag, char *txt) {
	struct s_object dummy;
	int distance = barwidth/2;

	dummy.xpos = 0;
	dummy.depth = 1;
	if(!src) src = &dummy;
	if(!dest) dest = &dummy;
	if(dest->newlyborn) {
		dest->newlyborn = 0;
		distance = dest->width / 2;
	}

	if(dotflag) {
		if(src->xpos < dest->xpos) {
			d_dottedarrow(src->xpos + src->depth*barwidth/2,
				curry - eventheight/2,
				dest->xpos - barwidth/2,
				curry - eventheight/2);
		} else {
			d_dottedarrow(src->xpos - barwidth/2,
				curry - eventheight/2,
				dest->xpos + dest->depth*barwidth/2,
				curry - eventheight/2);
		}
	} else {
		if(src->xpos < dest->xpos) {
			d_solidarrow(src->xpos + src->depth*barwidth/2,
				curry - eventheight/2,
				dest->xpos - distance,
				curry - eventheight/2);
		} else {
			d_solidarrow(src->xpos - barwidth/2,
				curry - eventheight/2,
				dest->xpos + dest->depth*distance,
				curry - eventheight/2);
		}
	}

	if(src->xpos < dest->xpos) {
		d_textlj(src->xpos + (src->depth+1)*barwidth/2,
			curry - eventheight/2 + textspacing,
			txt);
	} else {
		d_textrj(src->xpos - barwidth,
			curry - eventheight/2 + textspacing,
			txt);
	}
}

void drawobjhead(struct s_object *obj, int yoffs) {
	int i;

	i = obj->width/2;
	d_solid(obj->xpos - i, curry + yoffs,
		obj->xpos + i, curry + yoffs);
	d_solid(obj->xpos + i, curry + yoffs,
		obj->xpos + i, curry + yoffs - objheight);
	d_solid(obj->xpos + i, curry + yoffs - objheight,
		obj->xpos - i, curry + yoffs - objheight);
	d_solid(obj->xpos - i, curry + yoffs - objheight,
		obj->xpos - i, curry + yoffs);
	d_textcul(obj->xpos, curry + yoffs - objheight + textspacing + 3,
		obj->label);
}

void drawgetcalled(struct s_object *dest) {
	int i;

	switch(dest->state) {
		case OBJSTATE_INACTIVE:
			dest->state = OBJSTATE_ACTIVE;
			dest->depth = 1;
			d_dotted(dest->xpos, curry,
				dest->xpos, curry - eventheight/2);
			d_solid(dest->xpos - barwidth/2,
				curry - eventheight/2,
				dest->xpos + barwidth/2,
				curry - eventheight/2);
			d_solid(dest->xpos - barwidth/2,
				curry - eventheight/2,
				dest->xpos - barwidth/2,
				curry - eventheight);
			d_solid(dest->xpos + barwidth/2,
				curry - eventheight/2,
				dest->xpos + barwidth/2,
				curry - eventheight);
			break;
		case OBJSTATE_ACTIVE:
			d_solid(dest->xpos - barwidth/2,
				curry,
				dest->xpos - barwidth/2,
				curry - eventheight/2);
			for(i = 1; i < dest->depth; i++) {
				d_solid(dest->xpos + (i-1)*barwidth/2,
					curry,
					dest->xpos + (i-1)*barwidth/2,
					curry - eventheight/2);
			}
			d_solid(dest->xpos + dest->depth*barwidth/2,
				curry,
				dest->xpos + dest->depth*barwidth/2,
				curry - eventheight/2);
			dest->depth++;
			d_solid(dest->xpos - barwidth/2,
				curry - eventheight/2,
				dest->xpos - barwidth/2,
				curry - eventheight);
			for(i = 1; i < dest->depth; i++) {
				d_solid(dest->xpos + (i-1)*barwidth/2,
					curry - eventheight/2,
					dest->xpos + (i-1)*barwidth/2,
					curry - eventheight);
			}
			d_solid(dest->xpos + dest->depth*barwidth/2,
				curry - eventheight/2,
				dest->xpos + dest->depth*barwidth/2,
				curry - eventheight);
			d_solid(dest->xpos + dest->depth*barwidth/2 - barwidth,
				curry - eventheight/2,
				dest->xpos + dest->depth*barwidth/2,
				curry - eventheight/2);
			break;
		case OBJSTATE_UNBORN:
			dest->state = OBJSTATE_BEINGBORN1;
			dest->depth = 1;
			dest->newlyborn = 1;
			drawobjhead(dest, -eventheight/2);
			break;
		default:
			fprintf(stderr, "unimplemented state change\n");
			fprintf(stderr, "%d\n", dest->state);
			exit(3);
	}
}

void drawreturnage(struct s_object *dest) {
	int i;

	if(--dest->depth) {
		d_solid(dest->xpos - barwidth/2,
			curry,
			dest->xpos - barwidth/2,
			curry - eventheight/2);
		for(i = 1; i < (dest->depth+1); i++) {
			d_solid(dest->xpos + (i-1)*barwidth/2,
				curry,
				dest->xpos + (i-1)*barwidth/2,
				curry - eventheight/2);
		}
		d_solid(dest->xpos + (dest->depth+1)*barwidth/2,
			curry,
			dest->xpos + (dest->depth+1)*barwidth/2,
			curry - eventheight/2);
		d_solid(dest->xpos - barwidth/2,
			curry - eventheight/2,
			dest->xpos - barwidth/2,
			curry - eventheight);
		for(i = 1; i < dest->depth; i++) {
			d_solid(dest->xpos + (i-1)*barwidth/2,
				curry - eventheight/2,
				dest->xpos + (i-1)*barwidth/2,
				curry - eventheight);
		}
		d_solid(dest->xpos + dest->depth*barwidth/2,
			curry - eventheight/2,
			dest->xpos + dest->depth*barwidth/2,
			curry - eventheight);
		d_solid(dest->xpos + (dest->depth+1)*barwidth/2 - barwidth,
			curry - eventheight/2,
			dest->xpos + (dest->depth+1)*barwidth/2,
			curry - eventheight/2);
	} else {
		dest->state = OBJSTATE_INACTIVE;
		d_solid(dest->xpos - barwidth/2,
			curry,
			dest->xpos - barwidth/2,
			curry - eventheight/2);
		d_solid(dest->xpos + barwidth/2,
			curry,
			dest->xpos + barwidth/2,
			curry - eventheight/2);
		d_solid(dest->xpos - barwidth/2,
			curry - eventheight/2,
			dest->xpos + barwidth/2,
			curry - eventheight/2);
		d_dotted(dest->xpos, curry - eventheight/2,
			dest->xpos, curry - eventheight);
	}
}

void drawobjstate(struct s_object *obj, int height) {
	int i;

	switch(obj->state) {
		case OBJSTATE_INACTIVE:
			/*printf("draw dotted lifeline at %d\n", obj->xpos);*/
			d_dotted(obj->xpos, curry,
				obj->xpos, curry - height);
			break;
		case OBJSTATE_ACTIVE:
			d_solid(obj->xpos - barwidth/2,
				curry,
				obj->xpos - barwidth/2,
				curry - height);
			for(i = 1; i < obj->depth; i++) {
				d_solid(obj->xpos + (i-1)*barwidth/2,
					curry,
					obj->xpos + (i-1)*barwidth/2,
					curry - height);
			}
			d_solid(obj->xpos + obj->depth*barwidth/2,
				curry,
				obj->xpos + obj->depth*barwidth/2,
				curry - height);
			break;
		case OBJSTATE_BEINGBORN1:
			obj->state = OBJSTATE_BEINGBORN2;
			break;
		case OBJSTATE_BEINGBORN2:
			obj->state = OBJSTATE_ACTIVE;
			d_solid(obj->xpos - barwidth/2,
				curry - objheight + eventheight,
				obj->xpos - barwidth/2,
				curry - height);
			d_solid(obj->xpos + barwidth/2,
				curry - objheight + eventheight,
				obj->xpos + barwidth/2,
				curry - height);
			break;
		case OBJSTATE_UNBORN:
			break;
		default:
			fprintf(stderr, "unimplemented state\n");
	}
}

void drawgetdeleted(struct s_object *obj) {
	d_dotted(obj->xpos, curry, obj->xpos, curry - eventheight/2);
	d_extinction(obj->xpos, curry - eventheight/2);
	obj->state = OBJSTATE_UNBORN;
}

void drawaction(struct s_action *act, struct s_object *caller) {
	struct s_object *dest;
	int i;

	if(!act->msg) {
		/* this is a delay action */
		for(i = 0; i < nobjects; i++) {
			drawobjstate(objarray[i], eventheight/2);
		}
		curry -= eventheight/2;
		return;
	}

	dest = act->msg->obj;

	if(act->deletion) {
		/* this is a delete action */
		for(i = 0; i < nobjects; i++) {
			if(i == dest->id) {
				drawgetdeleted(dest);
			} else {
				drawobjstate(objarray[i], eventheight);
			}
		}
		if(*act->msg->msg) {
			drawarrow(caller, dest, 0, act->msg->msg);
		}
		curry -= eventheight;
		return;
	}

	for(i = 0; i < nobjects; i++) {
		if(i == dest->id) {
			drawgetcalled(dest);
		} else {
			drawobjstate(objarray[i], eventheight);
		}
	}

	if(caller && caller->id == dest->id) {
		/*printf("draw solid arrow-to-self\n");
		printf("  from %d to %d\n", dest->xpos, dest->xpos);
		printf("  with text \"%s\"\n", act->msg->msg);*/
		d_solid(caller->xpos + (caller->depth-1)*barwidth/2,
			curry - eventheight/4,
			caller->xpos + (caller->depth+3)*barwidth/2,
			curry - eventheight/4);
		d_solid(caller->xpos + (caller->depth+3)*barwidth/2,
			curry - eventheight/4,
			caller->xpos + (caller->depth+3)*barwidth/2,
			curry - eventheight/2);
		d_solidarrow(caller->xpos + (caller->depth+3)*barwidth/2,
			curry - eventheight/2,
			caller->xpos + caller->depth*barwidth/2,
			curry - eventheight/2);
		d_textlj(caller->xpos + caller->depth*barwidth/2,
			curry - eventheight/4 + textspacing,
			act->msg->msg);
	} else if(caller) {
		drawarrow(caller, dest, 0, act->msg->msg);
	} else {
		if(act->msg->msg[0]) {
			drawarrow(NULL, dest, 0, act->msg->msg);
		}
	}

	curry -= eventheight;

	drawactions(act->whathappen, dest);

	for(i = 0; i < nobjects; i++) {
		if(i == dest->id) {
			drawreturnage(dest);
		} else {
			drawobjstate(objarray[i], eventheight);
		}
	}

	if(caller && caller->id == dest->id) {
		d_dotted(dest->xpos + (dest->depth+1)*barwidth/2,
			curry - eventheight/2,
			dest->xpos + (dest->depth+4)*barwidth/2,
			curry - eventheight/2);
		d_dotted(dest->xpos + (dest->depth+4)*barwidth/2,
			curry - eventheight/2,
			dest->xpos + (dest->depth+4)*barwidth/2,
			curry - eventheight*3/4);
		d_dottedarrow(dest->xpos + (dest->depth+4)*barwidth/2,
			curry - eventheight*3/4,
			dest->xpos + dest->depth*barwidth/2,
			curry - eventheight*3/4);
		d_textlj(dest->xpos + (dest->depth+2)*barwidth/2,
			curry - eventheight/2 + textspacing,
			act->retval);
	} else if(caller) {
		drawarrow(dest, caller, 1, act->retval);
	} else {
		if(act->retval[0]) {
			drawarrow(dest, caller, 1, act->retval);
		}
	}

	curry -= eventheight;
}

void drawactions(struct s_action *act, struct s_object *caller) {

	while(act) {
		drawaction(act, caller);
		act = act->next;
	}
}

void recincheight(int *height, struct s_action *act) {
	if(act) {
		if(act->msg) {
			if(act->deletion) {
				*height += eventheight;
			} else {
				*height += 2*eventheight;
				recincheight(height, act->whathappen);
			}
		} else {
			*height += eventheight/2;
		}
		recincheight(height, act->next);
	}
}

void draw() {
	struct s_object *obj;
	struct s_text *txt;
	int i, maxxpos = 0, maxypos;

	for(obj = diagram->objects; obj; obj = obj->defnext) {
		if(!obj->width) obj->width = defobjwidth;
		i = obj->xpos + obj->width/2;
		if(i > maxxpos) maxxpos = i;
	}
	maxypos = objheight + eventheight/2;
	recincheight(&maxypos, diagram->seq);
	d_header(maxxpos + 1, maxypos + 1);
	curry = maxypos;

	nobjects = 0;
	for(obj = diagram->objects; obj; obj = obj->defnext) {
		/*printf("draw object:\n");
		printf("  label = %s\n", obj->label);
		printf("  xpos = %d\n", obj->xpos);*/
		if(obj->state != OBJSTATE_UNBORN)
			drawobjhead(obj, 0);
		nobjects++;
	}
	curry -= objheight;

	objarray = malloc(sizeof(struct s_object *) * nobjects);
	i = 0;
	for(obj = diagram->objects; obj; obj = obj->defnext) {
		obj->id = i;
		objarray[i++] = obj;
		drawobjstate(obj, eventheight/2);
	}

	curry -= eventheight/2;

	drawactions(diagram->seq, NULL);

	for(txt = diagram->texts; txt; txt = txt->next) {
		/*printf("draw text:\n");
		printf("  content = %s\n", txt->text);
		printf("  x,y = %d, %d\n", txt->xpos, txt->ypos);*/
		d_textlj(txt->xpos, txt->ypos, txt->text);
	}

	d_footer();
}
