#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "structs.h"
#include "protos.h"

#define min(a, b) ((a)<(b)?(a):(b))

int dotdelta = 10;
int arrowwidth = 6, arrowheight = 3;
int extinctionw = 18, extinctionh = 14;

void d_solid(int x1, int y1, int x2, int y2) {
	printf("newpath %d %d moveto %d %d lineto stroke\n",
		x1, y1, x2, y2);
}

void d_dotted(int x1, int y1, int x2, int y2) {
	int i;

	if(y1 == y2) {
		if(x1 > x2) { x1 ^= x2; x2 ^= x1; x1 ^= x2; }
		for(i = x1; i < x2; i += dotdelta) {
			d_solid(i, y1, min(i + dotdelta/2, x2), y1);
		}
	} else {
		if(y1 > y2) { y1 ^= y2; y2 ^= y1; y1 ^= y2; }
		for(i = y1; i < y2; i += dotdelta) {
			d_solid(x1, i, x1, min(i + dotdelta/2, y2));
		}
	}
}

void d_solidarrow(int x1, int y1, int x2, int y2) {
	d_solid(x1, y1, x2, y2);
	if(x1 < x2) {
		d_arrowheadr(x2, y2);
	} else {
		d_arrowheadl(x2, y2);
	}
}

void d_dottedarrow(int x1, int y1, int x2, int y2) {
	d_dotted(x1, y1, x2, y2);
	if(x1 < x2) {
		d_arrowheadr(x2, y2);
	} else {
		d_arrowheadl(x2, y2);
	}
}

void d_textlj(int x1, int y1, char *txt) {
	char *t;
	if(*txt) {
		t = psescape(txt);
		printf("newpath %d %d moveto (%s) show\n",
			x1, y1, t);
		free(t);
	}
}

void d_textrj(int x1, int y1, char *txt) {
	char *t;
	if(*txt) {
		t = psescape(txt);
		printf("newpath %d %d moveto (%s) dup stringwidth pop "
		       "neg 0 rmoveto show\n", x1, y1, t);
		free(t);
	}
}

void d_textcul(int x1, int y1, char *txt) {
	char *t;
	if(*txt) {
		t = psescape(txt);
		printf("newpath %d %d moveto (%s) dup stringwidth pop 2 "
		       "div neg 2 rmoveto show\n",
		       	x1, y1, t);
		printf("newpath %d %d moveto (%s) stringwidth pop dup 2 div "
		       "neg 0 rmoveto 0 rlineto stroke\n",
		       x1, y1, t);
		free(t);
	}
}

void d_header(int w, int h) {
	printf("%%!PS-Adobe-2.0 EPSF-3.0\n");
	printf("%%%%Creator: umlseq 1.6\n");
	printf("%%%%Pages: 1\n");
	printf("%%%%BoundingBox: 0 0 %d %d\n", w, h);
	printf("%%%%EndComments\n");
	printf("\n%% Umlseq was written by Linus Akesson.\n");
	printf("%% Visit http://www.df.lth.se/~lft for more information.\n");
	printf("%%%%Page: 1 1\n");
	printf("gsave\n");
	printf("0 setgray 1 setlinewidth\n");
	printf("/Times-Roman findfont 12 scalefont setfont\n");
	/*printf("1 -1 scale 0 %d translate\n", h);*/
}

void d_footer() {
	printf("showpage\n");
	printf("grestore\n");
	printf("%%%%Trailer:\n");
	printf("%%%%EOF\n");
}

char *psescape(char *in) {
	char *out, *p;

	out = p = (char *) malloc(strlen(in) * 2 + 2);
	while(*in) {
		switch(*in) {
			case 10:
				*p++ = '\\';
				*p++ = 'n';
				break;
			case '\\':
				*p++ = '\\';
				*p++ = '\\';
				break;
			case '(':
				*p++ = '\\';
				*p++ = '(';
				break;
			case ')':
				*p++ = '\\';
				*p++ = ')';
				break;
			default:
				*p++ = *in;
		}
		in++;
	}
	*p = 0;
	return out;
}

void d_arrowheadr(int x1, int y1) {
	printf("newpath %d %d moveto -%d -%d rlineto 0 %d rlineto "
	       "closepath fill\n",
	       x1, y1, arrowwidth, arrowheight, 2*arrowheight);
}

void d_arrowheadl(int x1, int y1) {
	printf("newpath %d %d moveto %d -%d rlineto 0 %d rlineto "
	       "closepath fill\n",
	       x1, y1, arrowwidth, arrowheight, 2*arrowheight);
}

void d_extinction(int x, int y) {
	printf("gsave 3 setlinewidth\n");
	d_solid(x - extinctionw/2, y - extinctionh/2,
		x + extinctionw/2, y + extinctionh/2);
	d_solid(x + extinctionw/2, y - extinctionh/2,
		x - extinctionw/2, y + extinctionh/2);
	printf("grestore\n");
}

